using System;
using SlatedGameToolkit.Framework.Exceptions;
using SlatedGameToolkit.Framework.Graphics;
using SlatedGameToolkit.Framework.Graphics.Textures;

namespace SlatedGameToolkit.Framework.Graphics
{
    public sealed class GLTexture2DSet : IDisposable
    {
        private bool disposed;
        public int Count { get; private set; }
        private GLGenTextures genTextures;
        private GLTexParameteri texParameteri;
        private GLTexParameterfv texParameterfv;
        private GLDeleteTextures deleteTextures;
        private GLTexImage2D texImage2D;
        private GLBindTexture bindTexture;
        private GLGenerateMipMap generateMipMap;
        private readonly UIntPtr[] handle;
        public GLTexture2DSet(TextureData[] textureDatas, bool linear = false, bool mipmap = false)
        {
            genTextures = GLFunctionUtils.RetrieveGLDelegate<GLGenTextures>("glGenTextures");
            deleteTextures = GLFunctionUtils.RetrieveGLDelegate<GLDeleteTextures>("glDeleteTextures");
            bindTexture = GLFunctionUtils.RetrieveGLDelegate<GLBindTexture>("glBindTexture");
            texParameteri = GLFunctionUtils.RetrieveGLDelegate<GLTexParameteri>("glTexParameteri");
            texParameterfv = GLFunctionUtils.RetrieveGLDelegate<GLTexParameterfv>("glTexParameterfv");
            texImage2D = GLFunctionUtils.RetrieveGLDelegate<GLTexImage2D>("glTexImage2D");
            generateMipMap = GLFunctionUtils.RetrieveGLDelegate<GLGenerateMipMap>("glGenerateMipMap");
            handle = new UIntPtr[textureDatas.Length];
            genTextures((uint)textureDatas.Length, handle);
            for (uint i = 0; i < textureDatas.Length; i++) {
                Setup(i, textureDatas[i]);
            }
            bindTexture(GLEnums.GL_TEXTURE_2D, UIntPtr.Zero);
            OpenGLErrorException.CheckGLErrorStatus();

            this.Count = textureDatas.Length;
        }

        private void Setup(uint index, TextureData textureData) {
            Bind(index);
            texImage2D(GLEnums.GL_TEXTURE_2D, 0, (int)GLEnums.GL_RGBA, textureData.width, textureData.height, 0, GLEnums.GL_RGBA, GLEnums.GL_UNSIGNED_BYTE, textureData.data);
            generateMipMap(GLEnums.GL_TEXTURE_2D);
        }

        /// <summary>
        /// Binds the given index's texture.
        /// Sets the properties given the index of the OpenGL 2D texture.
        /// </summary>
        /// <param name="index">The index of the texture.</param>
        /// <param name="linear">Whether or not to have linear filtering.</param>
        /// <param name="mipmap">Whether or not to use MipMap</param>
        public void SetProperties(uint index, GLEnums min, GLEnums mag) {
            Bind(index);
            texParameteri(GLEnums.GL_TEXTURE_2D, GLEnums.GL_TEXTURE_WRAP_S, (int)GLEnums.GL_CLAMP_TO_EDGE);
            texParameteri(GLEnums.GL_TEXTURE_2D, GLEnums.GL_TEXTURE_WRAP_T, (int)GLEnums.GL_CLAMP_TO_EDGE);
            texParameteri(GLEnums.GL_TEXTURE_2D, GLEnums.GL_TEXTURE_MIN_FILTER, (int) min);
            texParameteri(GLEnums.GL_TEXTURE_2D, GLEnums.GL_TEXTURE_MAG_FILTER, (int) mag);
        }

        /// <summary>
        /// Binds a GL texture to the active unit.
        /// </summary>
        /// <param name="index">The index of the texture in this set.</param>
        public void Bind(uint index) {
            bindTexture(GLEnums.GL_TEXTURE_2D, handle[index]);

        }

        public void Dispose()
        {
            if (disposed) return;
            this.disposed = true;
            deleteTextures((uint)handle.Length, handle);
            OpenGLErrorException.CheckGLErrorStatus();
            GC.SuppressFinalize(this);
        }
    }
}