using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.Threading;

namespace SlatedGameToolkit.Framework.Loader {
    public class AssetManager : IDisposable {
        private readonly object assetThreadLock = new object();
        private Thread thread;
        private ConcurrentQueue<ILoadable> loadables;
        private volatile bool load;
        public bool Complete {
            get {
                return loadables.Count == 0;
            }
        }
        private ConcurrentDictionary<string, IAssetUseable> assets;
        public ConcurrentDictionary<Type, PathModifier> PathModifiers {get; private set; }
        public AssetManager() {
            this.loadables = new ConcurrentQueue<ILoadable>();
            this.assets = new ConcurrentDictionary<string, IAssetUseable>();
            this.PathModifiers = new ConcurrentDictionary<Type, PathModifier>();
            thread = new Thread(Process);
            this.load = true;
            thread.Name = "Asset-Loader";
            thread.IsBackground = true;
            thread.Start();
        }

        /// <summary>
        /// Queue's the loadable for batched loading.
        /// The file name of the loadable is the one the loadable is saved under.
        /// </summary>
        /// <param name="loadable">The loadable.</param>
        public void QueueLoad(ILoadable loadable) {
            loadables.Enqueue(loadable);
        }

        private void Process() {
            lock (assetThreadLock)
            {
                ILoadable loadable;
                while (load) {
                    while (loadables.TryDequeue(out loadable)) {
                        Load(loadable);
                    }
                    Monitor.Pulse(loadable);                
                }
            }
        }

        /// <summary>
        /// Loads the loadable and stores it under the filename given to the loadable.
        /// </summary>
        /// <param name="loadable">The loadable to load.</param>
        public void Load(ILoadable loadable) {
            string name = loadable.FileName;
            assets[name] = loadable.Load(PathModifiers[loadable.GetType()]);
        }

        /// <summary>
        /// Loads the currently queued asynchronously.
        /// </summary>
        public void LoadQueued() {
            if (Monitor.TryEnter(assetThreadLock)) {
                Monitor.Pulse(assetThreadLock);
            }
        }

        /// <summary>
        /// Unloads the asset under the given name.
        /// This disposes the asset as well as removes it from the manager.
        /// </summary>
        /// <param name="name">The name of the asset to unload.</param>
        public void Unload(string name) {
            IAssetUseable asset;
            assets.TryRemove(name, out asset);
            asset.Dispose();
        }

        /// <summary>
        /// Removes all the assets.
        /// </summary>
        public void UnloadAll() {
            foreach (string name in assets.Keys)
            {
                Unload(name);
            }
        }

        /// <summary>
        /// Disposes of all the stored assets.
        /// </summary>
        public void Dispose()
        {
            load = false;
            UnloadAll();
        }

        ~AssetManager() {
            Dispose();
        }
    }
}